import fs from 'node:fs';
import path from 'node:path';
// @ts-expect-error
import barbe from 'barbe';
import * as k8s from '@kubernetes/client-node';
import { isTest } from '@terascope/utils';
const MAX_RETRIES = isTest ? 2 : 3;
const RETRY_DELAY = isTest ? 50 : 1000; // time in ms
const resourcePath = path.join(process.cwd(), './packages/teraslice/src/lib/cluster/services/cluster/backends/kubernetesV2/');
export function makeTemplate(folder, fileName) {
    const filePath = path.join(resourcePath, folder, `${fileName}.hbs`);
    const templateData = fs.readFileSync(filePath, 'utf-8');
    const templateKeys = ['{{', '}}'];
    return (config) => {
        if (folder !== 'jobs' && (config.exName === undefined || config.exUid === undefined)) {
            throw new Error(`K8s config requires ${config.exName === undefined ? 'exName' : 'exUid'} to create a ${folder} template`);
        }
        if (folder !== 'services' && config.dockerImage === undefined) {
            throw new Error(`K8s config requires a dockerImage to create a ${folder} template`);
        }
        const templated = barbe(templateData, templateKeys, config);
        return JSON.parse(templated);
    };
}
// Convert bytes to MB and reduce by 10%
export function getMaxOldSpace(memory) {
    return Math.round(0.9 * (memory / 1024 / 1024));
}
export function setMaxOldSpaceViaEnv(envArr, jobEnv, memory) {
    const envObj = {};
    if (memory && memory > -1) {
        // Set NODE_OPTIONS to override max-old-space-size
        const maxOldSpace = getMaxOldSpace(memory);
        envObj.NODE_OPTIONS = `--max-old-space-size=${maxOldSpace}`;
    }
    Object.assign(envObj, jobEnv);
    Object.entries(envObj).forEach(([name, value]) => {
        envArr.push({
            name,
            value
        });
    });
}
export function getRetryConfig() {
    return {
        retries: MAX_RETRIES,
        delay: RETRY_DELAY
    };
}
export function isDeployment(resource) {
    return resource instanceof k8s.V1Deployment;
}
export function isJob(resource) {
    return resource instanceof k8s.V1Job;
}
export function isPod(resource) {
    return resource instanceof k8s.V1Pod;
}
export function isReplicaSet(resource) {
    return resource instanceof k8s.V1ReplicaSet;
}
export function isService(resource) {
    return resource instanceof k8s.V1Service;
}
export function isTSDeployment(manifest) {
    return manifest instanceof k8s.V1Deployment
        && manifest.metadata?.labels !== undefined
        && manifest.metadata.name !== undefined
        && manifest.spec?.replicas !== undefined
        && manifest.spec.template.metadata?.labels !== undefined
        && manifest.spec.template.spec?.containers[0].volumeMounts !== undefined
        && manifest.spec.template.spec.volumes !== undefined;
}
export function isTSJob(manifest) {
    return manifest instanceof k8s.V1Job
        && manifest.metadata?.labels !== undefined
        && manifest.metadata.name !== undefined
        && manifest.spec?.template.metadata?.labels !== undefined
        && manifest.spec.template.spec?.containers[0].volumeMounts !== undefined
        && manifest.spec.template.spec.volumes !== undefined;
}
export function isTSPod(manifest) {
    return manifest instanceof k8s.V1Pod
        && manifest.metadata?.name !== undefined
        && manifest.status !== undefined;
}
export function isTSReplicaSet(manifest) {
    return manifest instanceof k8s.V1ReplicaSet
        && manifest.metadata?.name !== undefined
        && manifest.status !== undefined;
}
export function isTSService(manifest) {
    return manifest instanceof k8s.V1Service
        && manifest.metadata?.name !== undefined
        && manifest.spec?.selector !== undefined
        && manifest.spec.ports !== undefined;
}
export function convertToTSResource(resource) {
    if (isDeployment(resource) && isTSDeployment(resource)) {
        return resource;
    }
    if (isJob(resource) && isTSJob(resource)) {
        return resource;
    }
    if (isPod(resource) && isTSPod(resource)) {
        return resource;
    }
    if (isReplicaSet(resource) && isTSReplicaSet(resource)) {
        return resource;
    }
    if (isService(resource) && isTSService(resource)) {
        return resource;
    }
    throw new Error('K8sResource missing required field(s) to be converted to TSResource.');
}
export function isDeploymentList(manifest) {
    return manifest.kind === 'DeploymentList';
}
export function isJobList(manifest) {
    return manifest.kind === 'JobList';
}
export function isPodList(manifest) {
    return manifest.kind === 'PodList';
}
export function isReplicaSetList(manifest) {
    return manifest.kind === 'ReplicaSetList';
}
export function isServiceList(manifest) {
    return manifest.kind === 'ServiceList';
}
export function isTSDeploymentList(manifest) {
    return manifest.kind === 'DeploymentList'
        && (manifest.items[0] ? isTSDeployment(manifest.items[0]) : true);
}
export function isTSJobList(manifest) {
    return manifest.kind === 'JobList'
        && (manifest.items[0] ? isTSJob(manifest.items[0]) : true);
}
export function isTSPodList(manifest) {
    return manifest.kind === 'PodList'
        && (manifest.items[0] ? isTSPod(manifest.items[0]) : true);
}
export function isTSReplicaSetList(manifest) {
    return manifest.kind === 'ReplicaSetList'
        && (manifest.items[0] ? isTSReplicaSet(manifest.items[0]) : true);
}
export function isTSServiceList(manifest) {
    return manifest.kind === 'ServiceList'
        && (manifest.items[0] ? isTSService(manifest.items[0]) : true);
}
export function convertToTSResourceList(resourceList) {
    resourceList.items.map((resource) => convertToTSResource(resource));
    if (isDeploymentList(resourceList) && isTSDeploymentList(resourceList)) {
        return resourceList;
    }
    if (isJobList(resourceList) && isTSJobList(resourceList)) {
        return resourceList;
    }
    if (isPodList(resourceList) && isTSPodList(resourceList)) {
        return resourceList;
    }
    if (isReplicaSetList(resourceList) && isTSReplicaSetList(resourceList)) {
        return resourceList;
    }
    if (isServiceList(resourceList) && isTSServiceList(resourceList)) {
        return resourceList;
    }
    throw new Error('K8sResource missing required field(s) to be converted to TSResourceList.');
}
//# sourceMappingURL=utils.js.map