import { v4 as uuid } from 'uuid';
import { TSError, makeISODate } from '@terascope/utils';
import { makeLogger } from '../workers/helpers/terafoundation.js';
import { TerasliceElasticsearchStorage } from './backends/elasticsearch_store.js';
export class JobsStorage {
    backend;
    jobType;
    logger;
    constructor(context) {
        const logger = makeLogger(context, 'job_storage');
        const config = context.sysconfig.teraslice;
        const jobType = 'job';
        const indexName = `${config.name}__jobs`;
        const backendConfig = {
            context,
            indexName,
            recordType: 'job',
            idField: 'job_id',
            fullResponse: false,
            logRecord: false,
            storageName: 'jobs',
            logger
        };
        this.logger = logger;
        this.jobType = jobType;
        this.backend = new TerasliceElasticsearchStorage(backendConfig);
    }
    async initialize() {
        await this.backend.initialize();
        this.logger.info('job storage initialized');
    }
    async get(jobId) {
        const doc = await this.backend.get(jobId);
        return doc;
    }
    async search(query, from, size, sort, fields) {
        const results = await this.backend.search(query, from, size, sort, fields);
        return results;
    }
    async create(record) {
        const date = makeISODate();
        const doc = Object.assign({}, record, {
            job_id: uuid(),
            _context: this.jobType,
            _created: date,
            _updated: date
        });
        try {
            await this.backend.create(doc);
        }
        catch (err) {
            throw new TSError(err, {
                reason: 'Failure to create job'
            });
        }
        return doc;
    }
    async update(jobId, updateSpec) {
        // We want to save the whole job as it is posted, update api does partial doc updates
        const results = await this.backend.indexWithId(jobId, Object.assign({}, updateSpec, {
            job_id: jobId,
            _context: this.jobType,
            _updated: makeISODate()
        }));
        return results;
    }
    async remove(jobId) {
        return this.backend.remove(jobId);
    }
    async shutdown(forceShutdown) {
        this.logger.info('shutting down.');
        return this.backend.shutdown(forceShutdown);
    }
    verifyClient() {
        return this.backend.verifyClient();
    }
    async waitForClient() {
        return this.backend.waitForClient();
    }
}
//# sourceMappingURL=jobs.js.map