import { pRaceWithTimeout, logError, Queue } from '@terascope/utils';
import { makeLogger } from '../helpers/terafoundation.js';
export class RecoveryModule {
    logger;
    events;
    slicersToRecover;
    recoveryQueue;
    recoverComplete = true;
    isShutdown = false;
    autorecover;
    cleanupType;
    slicerID = 0;
    retryState = new Map();
    recoverExecution;
    exId;
    stateStore;
    timeout;
    constructor(context, executionContext) {
        const { exId } = executionContext;
        this.events = context.apis.foundation.getSystemEvents();
        this.slicersToRecover = executionContext.config.slicers;
        this.recoveryQueue = new Queue();
        this.cleanupType = executionContext.config.recovered_slice_type;
        this.recoverExecution = executionContext.config.recovered_execution;
        this.autorecover = Boolean(executionContext.config.autorecover);
        this.exId = exId;
        this.logger = makeLogger(context, 'execution_recovery');
        this.timeout = context.sysconfig.teraslice.shutdown_timeout;
    }
    initialize(stateStore) {
        this.stateStore = stateStore;
        this.events.on('slice:success', this._sliceComplete.bind(this));
        this.recoverComplete = false;
        // once we have fully recovered, clean up event listeners
        this.events.once('execution:recovery:complete', () => {
            this.events.removeListener('slice:success', this._sliceComplete.bind(this));
        });
    }
    // TODO: this is wrong
    _sliceComplete(sliceData) {
        this.retryState.set(sliceData.slice.slice_id, false);
    }
    _setId(slice) {
        this.retryState.set(slice.slice_id, true);
    }
    async _processIncompleteSlices(slicerID) {
        const slices = await this.stateStore.recoverSlices(this.recoverExecution, slicerID, this.cleanupType);
        for (const slice of slices) {
            this._setId(slice);
            this.recoveryQueue.enqueue(slice);
        }
        return slices.length;
    }
    _recoveryBatchCompleted() {
        return [...this.retryState.values()].every((v) => v === false);
    }
    _retryState() {
        return Object.fromEntries(this.retryState);
    }
    // TODO: refactor to use pwhile
    async _waitForRecoveryBatchCompletion() {
        return new Promise((resolve) => {
            const checkingBatch = setInterval(() => {
                if (this._recoveryBatchCompleted()) {
                    clearInterval(checkingBatch);
                    this.events.emit('execution:recovery:complete');
                    this.recoverComplete = true;
                    resolve(true);
                    return;
                }
                if (this.isShutdown) {
                    clearInterval(checkingBatch);
                    this.recoverComplete = false;
                    resolve(true);
                }
            }, 100);
        });
    }
    async handle() {
        if (this.recoverComplete) {
            return true;
        }
        const recoveredSlicesCount = await this._processIncompleteSlices(this.slicerID);
        if (recoveredSlicesCount === 0) {
            this.slicerID++;
            // all slicers have been recovered
            if (this.slicerID > this.slicersToRecover) {
                this.logger.warn(`recovered data for execution: ${this.exId} has successfully been enqueued`);
                await this._waitForRecoveryBatchCompletion();
                return true;
            }
        }
        await this._waitForRecoveryBatchCompletion();
        return false;
    }
    getSlice() {
        if (this.recoveryQueue.size() > 0) {
            return this.recoveryQueue.dequeue();
        }
        return;
    }
    getSlices(max = 1) {
        const count = max > this.sliceCount() ? this.sliceCount() : max;
        const slices = [];
        for (let i = 0; i < count; i++) {
            const slice = this.recoveryQueue.dequeue();
            if (!slice)
                return slices;
            slices.push(slice);
        }
        return slices;
    }
    sliceCount() {
        return this.recoveryQueue.size();
    }
    async shutdown() {
        let checkInterval;
        try {
            await pRaceWithTimeout(new Promise((resolve) => {
                checkInterval = setInterval(() => {
                    if (this.recoverComplete) {
                        resolve(true);
                    }
                }, 100);
            }), this.timeout, (err) => {
                logError(this.logger, err);
            });
        }
        finally {
            this.isShutdown = true;
            clearInterval(checkInterval);
        }
    }
    recoveryComplete() {
        return this.recoverComplete;
    }
    /**
     * Whether or not the execution will continue to process
     * slices after recovering.
     *
     * @returns {boolean}
    */
    exitAfterComplete() {
        if (this.autorecover)
            return false;
        if (!this.cleanupType)
            return false;
        return true;
    }
}
//# sourceMappingURL=recovery.js.map